/*
 * Copyright 2009 OW2 Chameleon
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.ow2.chameleon.sharedprefs.xml;

import java.io.File;
import java.io.IOException;

import org.junit.After;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;
import org.ow2.chameleon.sharedprefs.SharedPreferences;
import org.ow2.chameleon.sharedprefs.xml.SharedPreferencesServiceImpl;


public class SharedPreferencesServiceImplTest {

    private String m_location;
    
    private File getLocationFile() {
        return new File(m_location);
    }
    
    @Before
    public void setUp() {
        m_location = "target/tests";
    }
    
    @After
    public void tearDown() {
        // Delete all file
        delete(getLocationFile());
    }
    
    private void delete(File f) {
        if (f.isDirectory()) {
            for (File f2 : f.listFiles()) {
                delete(f2);
            }
            f.delete();
        } else {
            f.delete();
        }
    }
    
    @Test
    public void testSimple() throws IOException {
        SharedPreferencesServiceImpl svc = new SharedPreferencesServiceImpl();
        svc.setLocation(m_location);
        
        SharedPreferences prefs = svc.getSharedPreferences("simple");
        Assert.assertNotNull(prefs);
        Assert.assertTrue(prefs.getAll().isEmpty());
        Assert.assertNull(prefs.getString("foo", null));
        
        prefs.edit().putString("foo", "foo").commit();
        
        Assert.assertFalse(prefs.getAll().isEmpty());
        
        Assert.assertEquals("foo", prefs.getString("foo", null));
        
    }
    
    @Test
    public void testAllTypes() throws IOException {
        SharedPreferencesServiceImpl svc = new SharedPreferencesServiceImpl();
        svc.setLocation(m_location);
        
        SharedPreferences prefs = svc.getSharedPreferences("simple");
        Assert.assertNotNull(prefs);
        Assert.assertTrue(prefs.getAll().isEmpty());
        Assert.assertNull(prefs.getString("foo", null));
        Assert.assertEquals(prefs.getBoolean("bool", false), false);
        Assert.assertEquals(prefs.getFloat("float", 1.0f), 1.0f);
        Assert.assertEquals(prefs.getInt("int", 1), 1);
        Assert.assertEquals(prefs.getLong("long", 1l), 1l);
        
        prefs.edit()
            .putString("foo", "foo")
            .putBoolean("bool", true)
            .putFloat("float", 2.0f)
            .putInt("int", 2)
            .putLong("long", 2)
            .commit();
        
        Assert.assertFalse(prefs.getAll().isEmpty());
        
        Assert.assertEquals("foo", prefs.getString("foo", null));
        Assert.assertEquals(prefs.getBoolean("bool", false), true);
        Assert.assertEquals(prefs.getFloat("float", 2.0f), 2.0f);
        Assert.assertEquals(prefs.getInt("int", 1), 2);
        Assert.assertEquals(prefs.getLong("long", 1l), 2l);
        
    }
    
    @Test
    public void testReloads() throws IOException {
        SharedPreferencesServiceImpl svc = new SharedPreferencesServiceImpl();
        svc.setLocation("src/test/resources");
        
        SharedPreferences prefs = svc.getSharedPreferences("simple");
        
        Assert.assertFalse(prefs.getAll().isEmpty());
        
        Assert.assertEquals("foo", prefs.getString("foo", null));
        Assert.assertEquals(prefs.getBoolean("bool", false), true);
        Assert.assertEquals(prefs.getFloat("float", 2.0f), 2.0f);
        Assert.assertEquals(prefs.getInt("int", 1), 2);
        Assert.assertEquals(prefs.getLong("long", 1l), 2l);
        
    }
    
    @Test
    public void testNoCommit() throws IOException {
        SharedPreferencesServiceImpl svc = new SharedPreferencesServiceImpl();
        svc.setLocation(m_location);
        
        SharedPreferences prefs = svc.getSharedPreferences("simple");
        Assert.assertNotNull(prefs);
        Assert.assertTrue(prefs.getAll().isEmpty());
        Assert.assertNull(prefs.getString("foo", null));
        Assert.assertEquals(prefs.getBoolean("bool", false), false);
        Assert.assertEquals(prefs.getFloat("float", 1.0f), 1.0f);
        Assert.assertEquals(prefs.getInt("int", 1), 1);
        Assert.assertEquals(prefs.getLong("long", 1l), 1l);
        
        prefs.edit()
            .putString("foo", "foo")
            .putBoolean("bool", true)
            .putFloat("float", 2.0f)
            .putInt("int", 2)
            .putLong("long", 2); // NO COMMIT
        
        Assert.assertNotNull(prefs);
        Assert.assertTrue(prefs.getAll().isEmpty());
        Assert.assertNull(prefs.getString("foo", null));
        Assert.assertEquals(prefs.getBoolean("bool", false), false);
        Assert.assertEquals(prefs.getFloat("float", 1.0f), 1.0f);
        Assert.assertEquals(prefs.getInt("int", 1), 1);
        Assert.assertEquals(prefs.getLong("long", 1l), 1l);
        
        
    }
    
    @Test
    public void testThatWeGetTheSameSharedPreferenceObject() throws IOException {
        SharedPreferencesServiceImpl svc = new SharedPreferencesServiceImpl();
        svc.setLocation(m_location);
        
        SharedPreferences prefs1 = svc.getSharedPreferences("simple");
        SharedPreferences prefs2 = svc.getSharedPreferences("simple");
        Assert.assertNotNull(prefs1);
        Assert.assertNotNull(prefs2);
        Assert.assertTrue(prefs1.getAll().isEmpty());
        Assert.assertNull(prefs1.getString("foo", null));
        Assert.assertTrue(prefs2.getAll().isEmpty());
        Assert.assertNull(prefs2.getString("foo", null));
        
        Assert.assertEquals("Same objects", prefs1, prefs2);
        
        prefs1.edit().putString("foo", "foo").commit();
        
        SharedPreferences prefs3 = svc.getSharedPreferences("simple");

        // Prefs 2 updated.
        Assert.assertEquals("foo", prefs2.getString("foo", null));
        Assert.assertEquals("foo", prefs3.getString("foo", null));

        
    }
}
